/* out-next.c:  Functions for NeXT as target machine for GNU C compiler.

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* Make everything that used to go in the text section really go there.  */

int flag_no_mach_text_sections = 0;

#define INIT_PRAGMA_BUF_SIZE	80
#define PRAGMA_BUF_SIZE_INCR	40
#define OPT_STRCMP(opt) (!strncmp (opt, p, sizeof (opt)-1))

/* 1 if handle_pragma has been called yet.  */

static int pragma_initialized;

/* Initial setting of `optimize'.  */

static int initial_optimize_flag;

/* Called from check_newline via the macro HANDLE_PRAGMA.
   FINPUT is the source file input stream.  */

void
handle_pragma (finput)
     FILE *finput;
{
  char c;
  char *pragma_buf;
  int pragma_buf_size, i;
  char *p;
  extern int optimize, obey_regdecls, flag_writable_strings;

  /* Record initial setting of optimize flag, so we can restore it.  */
  if (!pragma_initialized)
    {
      pragma_initialized = 1;
      initial_optimize_flag = optimize;
    }

  /* Read the input line */
  pragma_buf_size = INIT_PRAGMA_BUF_SIZE;
  pragma_buf = (char *)xmalloc (pragma_buf_size);
  i = 0;
  while (c = getc (finput))
    {
      if (i == pragma_buf_size)
	{
	  pragma_buf_size += PRAGMA_BUF_SIZE_INCR;
	  pragma_buf = (char *)xrealloc (pragma_buf, pragma_buf_size);
	}
      if (c == '\n')
	{
	  ungetc (c, finput);
	  pragma_buf[i] = 0;
	  break;
	}
      pragma_buf[i] = c;
      i++;
    }
  p = pragma_buf;
  /* Skip white space */
  while ((*p == ' ') || (*p == '\t')) p++;

  if (OPT_STRCMP ("CC_OPT_ON"))
    {
      optimize = 1, obey_regdecls = 0;
      warning ("optimization turned on");
    }
  else if (OPT_STRCMP ("CC_OPT_OFF"))
    {
      optimize = 0, obey_regdecls = 1;
      warning ("optimization turned off");
    }
  else if (OPT_STRCMP ("CC_OPT_RESTORE"))
    {
      extern int initial_optimize_flag;

      if (optimize != initial_optimize_flag)
	{
	  if (initial_optimize_flag)
	    obey_regdecls = 0;
	  else
	    obey_regdecls = 1;
	  optimize = initial_optimize_flag;
	}
      warning ("optimization level restored");
    }
  else if (OPT_STRCMP ("CC_WRITABLE_STRINGS"))
    flag_writable_strings = 1;
  else if (OPT_STRCMP ("CC_NON_WRITABLE_STRINGS"))
    flag_writable_strings = 0;
  else if (OPT_STRCMP("CC_NO_MACH_TEXT_SECTIONS"))
    flag_no_mach_text_sections = 1;
  /* Ignore the rest of the line. */
  free (pragma_buf);
}
